@@ -16,17 +16,40 @@
 
 #include <dst/dst.h>
 
-#define RRTYPE_KEY_ATTRIBUTES (0)
+#define RRTYPE_KEY_ATTRIBUTES \
+	(DNS_RDATATYPEATTR_ATCNAME | DNS_RDATATYPEATTR_ZONECUTAUTH)
 
-static inline isc_result_t
-generic_fromtext_key(ARGS_FROMTEXT) {
-	isc_result_t result;
-	isc_token_t token;
-	dns_secalg_t alg;
+/*
+ * RFC 2535 section 3.1.2 says that if bits 0-1 of the Flags field are
+ * both set, it means there is no key information and the RR stops after
+ * the algorithm octet.  However, this only applies to KEY records, as
+ * indicated by the specifications of the RR types based on KEY:
+ *
+ *     CDNSKEY - RFC 7344
+ *     DNSKEY - RFC 4034
+ *     RKEY - draft-reid-dnsext-rkey-00
+ */
+static inline bool
+generic_key_nokey(dns_rdatatype_t type, unsigned int flags)
+{
+	switch (type) {
+	case dns_rdatatype_cdnskey:
+	case dns_rdatatype_dnskey:
+	case dns_rdatatype_rkey:
+		return (false);
+	case dns_rdatatype_key:
+	default:
+		return ((flags & DNS_KEYFLAG_TYPEMASK) == DNS_KEYTYPE_NOKEY);
+	}
+}
+
+static inline isc_result_t generic_fromtext_key(ARGS_FROMTEXT)
+{
+	isc_token_t    token;
+	dns_secalg_t   alg;
 	dns_secproto_t proto;
 	dns_keyflags_t flags;
 
-	UNUSED(type);
 	UNUSED(rdclass);
 	UNUSED(origin);
 	UNUSED(options);
@@ -34,46 +57,42 @@ generic_fromtext_key(ARGS_FROMTEXT) {
 
 	/* flags */
 	RETERR(isc_lex_getmastertoken(lexer, &token, isc_tokentype_string,
-				      ISC_FALSE));
+				      false));
 	RETTOK(dns_keyflags_fromtext(&flags, &token.value.as_textregion));
+	if (type == dns_rdatatype_rkey && flags != 0U) {
+		RETTOK(DNS_R_FORMERR);
+	}
 	RETERR(uint16_tobuffer(flags, target));
 
 	/* protocol */
 	RETERR(isc_lex_getmastertoken(lexer, &token, isc_tokentype_string,
-				      ISC_FALSE));
+				      false));
 	RETTOK(dns_secproto_fromtext(&proto, &token.value.as_textregion));
 	RETERR(mem_tobuffer(target, &proto, 1));
 
 	/* algorithm */
 	RETERR(isc_lex_getmastertoken(lexer, &token, isc_tokentype_string,
-				      ISC_FALSE));
+				      false));
 	RETTOK(dns_secalg_fromtext(&alg, &token.value.as_textregion));
 	RETERR(mem_tobuffer(target, &alg, 1));
 
 	/* No Key? */
-	if ((flags & 0xc000) == 0xc000)
+	if (generic_key_nokey(type, flags)) {
 		return (ISC_R_SUCCESS);
+	}
 
-	result = isc_base64_tobuffer(lexer, target, -1);
-	if (result != ISC_R_SUCCESS)
-		return (result);
-
-	/* Ensure there's at least enough data to compute a key ID for MD5 */
-	if (alg == DST_ALG_RSAMD5 && isc_buffer_usedlength(target) < 7)
-		return (ISC_R_UNEXPECTEDEND);
-
-	return (ISC_R_SUCCESS);
+	return (isc_base64_tobuffer(lexer, target, -2));
 }
 
-static inline isc_result_t
-generic_totext_key(ARGS_TOTEXT) {
-	isc_region_t sr;
-	char buf[sizeof("[key id = 64000]")];
-	unsigned int flags;
+static inline isc_result_t generic_totext_key(ARGS_TOTEXT)
+{
+	isc_region_t  sr;
+	char	      buf[sizeof("[key id = 64000]")];
+	unsigned int  flags;
 	unsigned char algorithm;
-	char algbuf[DNS_NAME_FORMATSIZE];
-	const char *keyinfo;
-	isc_region_t tmpr;
+	char	      algbuf[DNS_NAME_FORMATSIZE];
+	const char *  keyinfo;
+	isc_region_t  tmpr;
 
 	REQUIRE(rdata->length != 0);
 
@@ -93,7 +112,6 @@ generic_totext_key(ARGS_TOTEXT) {
 	} else
 		keyinfo = "ZSK";
 
-
 	/* protocol */
 	snprintf(buf, sizeof(buf), "%u", sr.base[0]);
 	isc_region_consume(&sr, 1);
@@ -107,17 +125,18 @@ generic_totext_key(ARGS_TOTEXT) {
 	RETERR(str_totext(buf, target));
 
 	/* No Key? */
-	if ((flags & 0xc000) == 0xc000)
+	if (generic_key_nokey(rdata->type, flags)) {
 		return (ISC_R_SUCCESS);
+	}
 
 	if ((tctx->flags & DNS_STYLEFLAG_RRCOMMENT) != 0 &&
-	     algorithm == DNS_KEYALG_PRIVATEDNS) {
+	    algorithm == DNS_KEYALG_PRIVATEDNS) {
 		dns_name_t name;
 		dns_name_init(&name, NULL);
 		dns_name_fromregion(&name, &sr);
 		dns_name_format(&name, algbuf, sizeof(algbuf));
 	} else {
-		dns_secalg_format((dns_secalg_t) algorithm, algbuf,
+		dns_secalg_format((dns_secalg_t)algorithm, algbuf,
 				  sizeof(algbuf));
 	}
 
@@ -127,7 +146,7 @@ generic_totext_key(ARGS_TOTEXT) {
 	RETERR(str_totext(tctx->linebreak, target));
 
 	if ((tctx->flags & DNS_STYLEFLAG_NOCRYPTO) == 0) {
-		if (tctx->width == 0)   /* No splitting */
+		if (tctx->width == 0) /* No splitting */
 			RETERR(isc_base64_totext(&sr, 60, "", target));
 		else
 			RETERR(isc_base64_totext(&sr, tctx->width - 2,
@@ -135,7 +154,7 @@ generic_totext_key(ARGS_TOTEXT) {
 	} else {
 		dns_rdata_toregion(rdata, &tmpr);
 		snprintf(buf, sizeof(buf), "[key id = %u]",
-			 dst_region_computeid(&tmpr, algorithm));
+			 dst_region_computeid(&tmpr));
 		RETERR(str_totext(buf, target));
 	}
 
@@ -148,7 +167,6 @@ generic_totext_key(ARGS_TOTEXT) {
 		RETERR(str_totext(")", target));
 
 	if ((tctx->flags & DNS_STYLEFLAG_RRCOMMENT) != 0) {
-
 		if (rdata->type == dns_rdatatype_dnskey ||
 		    rdata->type == dns_rdatatype_cdnskey) {
 			RETERR(str_totext(" ; ", target));
@@ -158,32 +176,44 @@ generic_totext_key(ARGS_TOTEXT) {
 		RETERR(str_totext(algbuf, target));
 		RETERR(str_totext(" ; key id = ", target));
 		dns_rdata_toregion(rdata, &tmpr);
-		snprintf(buf, sizeof(buf), "%u",
-			 dst_region_computeid(&tmpr, algorithm));
+		snprintf(buf, sizeof(buf), "%u", dst_region_computeid(&tmpr));
 		RETERR(str_totext(buf, target));
 	}
 	return (ISC_R_SUCCESS);
 }
 
-static inline isc_result_t
-generic_fromwire_key(ARGS_FROMWIRE) {
+static inline isc_result_t generic_fromwire_key(ARGS_FROMWIRE)
+{
 	unsigned char algorithm;
-	isc_region_t sr;
+	uint16_t      flags;
+	isc_region_t  sr;
 
-	UNUSED(type);
 	UNUSED(rdclass);
 	UNUSED(dctx);
 	UNUSED(options);
 
 	isc_buffer_activeregion(source, &sr);
-	if (sr.length < 4)
+	if (sr.length < 4) {
 		return (ISC_R_UNEXPECTEDEND);
+	}
+	flags = (sr.base[0] << 8) | sr.base[1];
+
+	if (type == dns_rdatatype_rkey && flags != 0U) {
+		return (DNS_R_FORMERR);
+	}
 
 	algorithm = sr.base[3];
 	RETERR(mem_tobuffer(target, sr.base, 4));
 	isc_region_consume(&sr, 4);
 	isc_buffer_forward(source, 4);
 
+	if (generic_key_nokey(type, flags)) {
+		return (ISC_R_SUCCESS);
+	}
+	if (sr.length == 0) {
+		return (ISC_R_UNEXPECTEDEND);
+	}
+
 	if (algorithm == DNS_KEYALG_PRIVATEDNS) {
 		dns_name_t name;
 		dns_decompress_setmethods(dctx, DNS_COMPRESS_NONE);
@@ -191,48 +221,37 @@ generic_fromwire_key(ARGS_FROMWIRE) {
 		RETERR(dns_name_fromwire(&name, source, dctx, options, target));
 	}
 
-	/*
-	 * RSAMD5 computes key ID differently from other
-	 * algorithms: we need to ensure there's enough data
-	 * present for the computation
-	 */
-	if (algorithm == DST_ALG_RSAMD5 && sr.length < 3)
-		return (ISC_R_UNEXPECTEDEND);
-
 	isc_buffer_activeregion(source, &sr);
 	isc_buffer_forward(source, sr.length);
 	return (mem_tobuffer(target, sr.base, sr.length));
 }
 
-static inline isc_result_t
-fromtext_key(ARGS_FROMTEXT) {
-
+static inline isc_result_t fromtext_key(ARGS_FROMTEXT)
+{
 	REQUIRE(type == dns_rdatatype_key);
 
-	return (generic_fromtext_key(rdclass, type, lexer, origin,
-				     options, target, callbacks));
+	return (generic_fromtext_key(rdclass, type, lexer, origin, options,
+				     target, callbacks));
 }
 
-static inline isc_result_t
-totext_key(ARGS_TOTEXT) {
-
+static inline isc_result_t totext_key(ARGS_TOTEXT)
+{
 	REQUIRE(rdata != NULL);
 	REQUIRE(rdata->type == dns_rdatatype_key);
 
 	return (generic_totext_key(rdata, tctx, target));
 }
 
-static inline isc_result_t
-fromwire_key(ARGS_FROMWIRE) {
-
+static inline isc_result_t fromwire_key(ARGS_FROMWIRE)
+{
 	REQUIRE(type == dns_rdatatype_key);
 
-	return (generic_fromwire_key(rdclass, type, source, dctx,
-				     options, target));
+	return (generic_fromwire_key(rdclass, type, source, dctx, options,
+				     target));
 }
 
-static inline isc_result_t
-towire_key(ARGS_TOWIRE) {
+static inline isc_result_t towire_key(ARGS_TOWIRE)
+{
 	isc_region_t sr;
 
 	REQUIRE(rdata != NULL);
@@ -245,8 +264,8 @@ towire_key(ARGS_TOWIRE) {
 	return (mem_tobuffer(target, sr.base, sr.length));
 }
 
-static inline int
-compare_key(ARGS_COMPARE) {
+static inline int compare_key(ARGS_COMPARE)
+{
 	isc_region_t r1;
 	isc_region_t r2;
 
@@ -263,8 +282,8 @@ compare_key(ARGS_COMPARE) {
 	return (isc_region_compare(&r1, &r2));
 }
 
-static inline isc_result_t
-generic_fromstruct_key(ARGS_FROMSTRUCT) {
+static inline isc_result_t generic_fromstruct_key(ARGS_FROMSTRUCT)
+{
 	dns_rdata_key_t *key = source;
 
 	REQUIRE(key != NULL);
@@ -274,6 +293,10 @@ generic_fromstruct_key(ARGS_FROMSTRUCT) {
 	UNUSED(type);
 	UNUSED(rdclass);
 
+	if (type == dns_rdatatype_rkey) {
+		INSIST(key->flags == 0U);
+	}
+
 	/* Flags */
 	RETERR(uint16_tobuffer(key->flags, target));
 
@@ -287,12 +310,12 @@ generic_fromstruct_key(ARGS_FROMSTRUCT) {
 	return (mem_tobuffer(target, key->data, key->datalen));
 }
 
-static inline isc_result_t
-generic_tostruct_key(ARGS_TOSTRUCT) {
+static inline isc_result_t generic_tostruct_key(ARGS_TOSTRUCT)
+{
 	dns_rdata_key_t *key = target;
-	isc_region_t sr;
+	isc_region_t	 sr;
 
-	REQUIRE(rdata != NULL);
+	REQUIRE(key != NULL);
 	REQUIRE(rdata->length != 0);
 
 	REQUIRE(key != NULL);
@@ -330,9 +353,9 @@ generic_tostruct_key(ARGS_TOSTRUCT) {
 	return (ISC_R_SUCCESS);
 }
 
-static inline void
-generic_freestruct_key(ARGS_FREESTRUCT) {
-	dns_rdata_key_t *key = (dns_rdata_key_t *) source;
+static inline void generic_freestruct_key(ARGS_FREESTRUCT)
+{
+	dns_rdata_key_t *key = (dns_rdata_key_t *)source;
 
 	REQUIRE(key != NULL);
 
@@ -344,16 +367,15 @@ generic_freestruct_key(ARGS_FREESTRUCT) {
 	key->mctx = NULL;
 }
 
-static inline isc_result_t
-fromstruct_key(ARGS_FROMSTRUCT) {
-
+static inline isc_result_t fromstruct_key(ARGS_FROMSTRUCT)
+{
 	REQUIRE(type == dns_rdatatype_key);
 
 	return (generic_fromstruct_key(rdclass, type, source, target));
 }
 
-static inline isc_result_t
-tostruct_key(ARGS_TOSTRUCT) {
+static inline isc_result_t tostruct_key(ARGS_TOSTRUCT)
+{
 	dns_rdata_key_t *key = target;
 
 	REQUIRE(key != NULL);
@@ -367,9 +389,9 @@ tostruct_key(ARGS_TOSTRUCT) {
 	return (generic_tostruct_key(rdata, target, mctx));
 }
 
-static inline void
-freestruct_key(ARGS_FREESTRUCT) {
-	dns_rdata_key_t *key = (dns_rdata_key_t *) source;
+static inline void freestruct_key(ARGS_FREESTRUCT)
+{
+	dns_rdata_key_t *key = (dns_rdata_key_t *)source;
 
 	REQUIRE(key != NULL);
 	REQUIRE(key->common.rdtype == dns_rdatatype_key);
@@ -377,9 +399,8 @@ freestruct_key(ARGS_FREESTRUCT) {
 	generic_freestruct_key(source);
 }
 
-static inline isc_result_t
-additionaldata_key(ARGS_ADDLDATA) {
-
+static inline isc_result_t additionaldata_key(ARGS_ADDLDATA)
+{
 	REQUIRE(rdata != NULL);
 	REQUIRE(rdata->type == dns_rdatatype_key);
 
@@ -390,8 +411,8 @@ additionaldata_key(ARGS_ADDLDATA) {
 	return (ISC_R_SUCCESS);
 }
 
-static inline isc_result_t
-digest_key(ARGS_DIGEST) {
+static inline isc_result_t digest_key(ARGS_DIGEST)
+{
 	isc_region_t r;
 
 	REQUIRE(rdata != NULL);
@@ -402,9 +423,8 @@ digest_key(ARGS_DIGEST) {
 	return ((digest)(arg, &r));
 }
 
-static inline isc_boolean_t
-checkowner_key(ARGS_CHECKOWNER) {
-
+static inline bool checkowner_key(ARGS_CHECKOWNER)
+{
 	REQUIRE(type == dns_rdatatype_key);
 
 	UNUSED(name);
@@ -412,12 +432,11 @@ checkowner_key(ARGS_CHECKOWNER) {
 	UNUSED(rdclass);
 	UNUSED(wildcard);
 
-	return (ISC_TRUE);
+	return (true);
 }
 
-static inline isc_boolean_t
-checknames_key(ARGS_CHECKNAMES) {
-
+static inline bool checknames_key(ARGS_CHECKNAMES)
+{
 	REQUIRE(rdata != NULL);
 	REQUIRE(rdata->type == dns_rdatatype_key);
 
@@ -425,12 +444,12 @@ checknames_key(ARGS_CHECKNAMES) {
 	UNUSED(owner);
 	UNUSED(bad);
 
-	return (ISC_TRUE);
+	return (true);
 }
 
-static inline int
-casecompare_key(ARGS_COMPARE) {
+static inline int casecompare_key(ARGS_COMPARE)
+{
 	return (compare_key(rdata1, rdata2));
 }
 
-#endif	/* RDATA_GENERIC_KEY_25_C */
+#endif /* RDATA_GENERIC_KEY_25_C */
